"""Template store."""

from typing import Dict, Optional, Sequence

from dreamsboard.engine.schema import BaseNode
from dreamsboard.engine.generate.code_generate import (
    CodeGenerator
)

from dreamsboard.engine.storage.template_store.types import BaseTemplateStore, RefTemplateInfo
from dreamsboard.engine.storage.template_store.utils import template_to_json, json_to_template
from dreamsboard.engine.storage.kvstore.types import BaseKVStore

DEFAULT_NAMESPACE = "template_store"


class KVTemplateStore(BaseTemplateStore):
    """Template (Node) store.

    NOTE: at the moment, this store is primarily used to store Node objects.
    Each node will be assigned an ID.

    The same templatestore can be reused across index structures. This
    allows you to reuse the same storage for multiple index structures;
    otherwise, each index would create a templatestore under the hood.

    .. code-block:: python
        nodes = SimpleNodeParser.get_nodes_from_templates()
        templatestore = SimpleTemplateStore()
        templatestore.add_templates(nodes)
        storage_context = StorageContext.from_defaults(templatestore=templatestore)

        summary_index = SummaryIndex(nodes, storage_context=storage_context)
        vector_index = VectorStoreIndex(nodes, storage_context=storage_context)
        keyword_table_index = SimpleKeywordTableIndex(
            nodes,
            storage_context=storage_context
        )

    This will use the same templatestore for multiple index structures.

    Args:
        kvstore (BaseKVStore): key-value store
        namespace (str): namespace for the templatestore

    """

    def __init__(
        self,
        kvstore: BaseKVStore,
        namespace: Optional[str] = None,
    ) -> None:
        """Init a KVTemplateStore."""
        self._kvstore = kvstore
        self._namespace = namespace or DEFAULT_NAMESPACE
        self._node_collection = f"{self._namespace}/data"
        self._ref_template_collection = f"{self._namespace}/ref_template_info"
        self._metadata_collection = f"{self._namespace}/metadata"

    @property
    def templates(self) -> Dict[str, BaseNode]:
        """Get all templates.

        Returns:
            Dict[str, BaseTemplate]: templates

        """
        json_dict = self._kvstore.get_all(collection=self._node_collection)
        return {key: json_to_template(json) for key, json in json_dict.items()}

    def add_templates(
        self, nodes: Sequence[BaseNode], allow_update: bool = True
    ) -> None:
        """Add a template to the store.

        Args:
            templates (List[BaseTemplate]): templates
            allow_update (bool): allow update of templatestore from template
            :param allow_update:
            :param nodes:

        """
        for node in nodes:
            # NOTE: template could already exist in the store, but we overwrite it
            if not allow_update and self.template_exists(node.node_id):
                raise ValueError(
                    f"node_id {node.node_id} already exists. "
                    "Set allow_update to True to overwrite."
                )
            node_key = node.node_id
            data = template_to_json(node)
            self._kvstore.put(node_key, data, collection=self._node_collection)

            # update template_collection if needed
            metadata = {"template_hash": node.hash}
            if isinstance(node, CodeGenerator) and node.ref_template_id is not None:
                ref_template_info = self.get_ref_template_info(node.ref_template_id) or RefTemplateInfo()
                ref_template_info.node_ids.append(node.node_id)
                if not ref_template_info.metadata:
                    ref_template_info.metadata = node.metadata or {}
                self._kvstore.put(
                    node.ref_template_id,
                    ref_template_info.to_dict(),
                    collection=self._ref_template_collection,
                )

                # update metadata with map
                metadata["ref_template_id"] = node.ref_template_id
                self._kvstore.put(
                    node_key, metadata, collection=self._metadata_collection
                )
            else:
                self._kvstore.put(
                    node_key, metadata, collection=self._metadata_collection
                )

    def get_template(self, template_id: str, raise_error: bool = True) -> Optional[BaseNode]:
        """Get a template from the store.

        Args:
            template_id (str): template id
            raise_error (bool): raise error if template_id not found

        """
        json = self._kvstore.get(template_id, collection=self._node_collection)
        if json is None:
            if raise_error:
                raise ValueError(f"template_id {template_id} not found.")
            else:
                return None
        return json_to_template(json)

    def get_ref_template_info(self, ref_template_id: str) -> Optional[RefTemplateInfo]:
        """Get the RefTemplateInfo for a given ref_template_id."""
        ref_template_info = self._kvstore.get(
            ref_template_id, collection=self._ref_template_collection
        )
        if not ref_template_info:
            return None

        # TODO: deprecated legacy support
        if "template_ids" in ref_template_info:
            ref_template_info["node_ids"] = ref_template_info.get("template_ids", [])
            ref_template_info.pop("template_ids")

            ref_template_info["metadata"] = ref_template_info.get("extra_info", {})
            ref_template_info.pop("extra_info")

        return RefTemplateInfo(**ref_template_info)

    def get_all_ref_template_info(self) -> Optional[Dict[str, RefTemplateInfo]]:
        """Get a mapping of ref_template_id -> RefTemplateInfo for all ingested templates."""
        ref_template_infos = self._kvstore.get_all(collection=self._ref_template_collection)
        if ref_template_infos is None:
            return None

        # TODO: deprecated legacy support
        all_ref_template_infos = {}
        for template_id, ref_template_info in ref_template_infos.items():
            if "template_ids" in ref_template_info:
                ref_template_info["node_ids"] = ref_template_info.get("template_ids", [])
                ref_template_info.pop("template_ids")

                ref_template_info["metadata"] = ref_template_info.get("extra_info", {})
                ref_template_info.pop("extra_info")
                all_ref_template_infos[template_id] = RefTemplateInfo(**ref_template_info)

        return all_ref_template_infos

    def ref_template_exists(self, ref_template_id: str) -> bool:
        """Check if a ref_template_id has been ingested."""
        return self.get_ref_template_info(ref_template_id) is not None

    def template_exists(self, template_id: str) -> bool:
        """Check if template exists."""
        return self._kvstore.get(template_id, self._node_collection) is not None

    def _remove_ref_template_node(self, template_id: str) -> None:
        """Helper function to remove node template_id from ref_template_collection."""
        metadata = self._kvstore.get(template_id, collection=self._metadata_collection)
        if metadata is None:
            return

        ref_template_id = metadata.get("ref_template_id", None)

        if ref_template_id is None:
            return

        ref_template_info = self._kvstore.get(
            ref_template_id, collection=self._ref_template_collection
        )

        if ref_template_info is not None:
            ref_template_obj = RefTemplateInfo(**ref_template_info)

            ref_template_obj.node_ids.remove(template_id)

            # delete ref_template from collection if it has no more template_ids
            if len(ref_template_obj.node_ids) > 0:
                self._kvstore.put(
                    ref_template_id,
                    ref_template_obj.to_dict(),
                    collection=self._ref_template_collection,
                )

            self._kvstore.delete(ref_template_id, collection=self._metadata_collection)

    def delete_template(
        self, template_id: str, raise_error: bool = True, remove_ref_template_node: bool = True
    ) -> None:
        """Delete a template from the store."""
        if remove_ref_template_node:
            self._remove_ref_template_node(template_id)

        delete_success = self._kvstore.delete(template_id, collection=self._node_collection)
        _ = self._kvstore.delete(template_id, collection=self._metadata_collection)

        if not delete_success and raise_error:
            raise ValueError(f"template_id {template_id} not found.")

    def delete_ref_template(self, ref_template_id: str, raise_error: bool = True) -> None:
        """Delete a ref_template and all it's associated nodes."""
        ref_template_info = self.get_ref_template_info(ref_template_id)
        if ref_template_info is None:
            if raise_error:
                raise ValueError(f"ref_template_id {ref_template_id} not found.")
            else:
                return

        for template_id in ref_template_info.node_ids:
            self.delete_template(template_id, raise_error=False, remove_ref_template_node=False)

        self._kvstore.delete(ref_template_id, collection=self._metadata_collection)
        self._kvstore.delete(ref_template_id, collection=self._ref_template_collection)

    def set_template_hash(self, template_id: str, template_hash: str) -> None:
        """Set the hash for a given template_id."""
        metadata = {"template_hash": template_hash}
        self._kvstore.put(template_id, metadata, collection=self._metadata_collection)

    def get_template_hash(self, template_id: str) -> Optional[str]:
        """Get the stored hash for a template, if it exists."""
        metadata = self._kvstore.get(template_id, collection=self._metadata_collection)
        if metadata is not None:
            return metadata.get("template_hash", None)
        else:
            return None



"""Index registry."""

from typing import Dict, Type

from dreamsboard.engine.data_structs.struct_type import IndexStructType
from dreamsboard.engine.engine_builder import BaseEngineBuilder, CodeGeneratorBuilder

INDEX_STRUCT_TYPE_TO_INDEX_CLASS: Dict[IndexStructType, Type[BaseEngineBuilder]] = {
    IndexStructType.CODE_GENERATOR: CodeGeneratorBuilder,
}



from __future__ import annotations

from abc import abstractmethod, ABC
from typing import Any, Optional, Dict

import logging

from pydantic import Field

from dreamsboard.engine.schema import BaseNode, ObjectTemplateType
from dreamsboard.templates import get_template_path

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# 控制台打印
handler = logging.StreamHandler()
handler.setLevel(logging.INFO)

logger.addHandler(handler)


class DreamsPersonalityNode(BaseNode, ABC):
    """心里分析信息节点"""
    story_scenario_context: Optional[str] = Field(
        default="", description="开放问题 故事情境"
    )
    scene_monologue_context: Optional[str] = Field(
        default="", description="开放问题 故事场景"
    )
    user_id: Optional[str] = Field(
        default="", description="开放问题user_id"
    )
    scene_content: Optional[str] = Field(
        default="", description="开放问题 文本内容"
    )
    story_board_summary_context: Optional[str] = Field(
        default="", description="开放问题 人物对话"
    )
    dreams_guidance_context: Optional[str] = Field(
        default="", description="开放性问题dreams_guidance_context"
    )
    evolutionary_step: Optional[str] = Field(
        default="", description="性格信息 剧情总结",
    )
    dreams_personality_context: Optional[str] = Field(
        default="", description="性格信息dreams_personality_context",
    )
    ref_analysis_id: Optional[str] = Field(
        default="", description="ref_analysis_id",
    )

    def __init__(self,
                 story_scenario_context: str = None,
                 scene_monologue_context: str = None,
                 user_id: str = None,
                 scene_content: str = None,
                 story_board_summary_context: str = None,
                 dreams_guidance_context: str = None,
                 evolutionary_step: str = None,
                 dreams_personality_context=None,
                 ref_analysis_id: str = None,
                 **kwargs):
        super().__init__(**kwargs)
        self.story_scenario_context = story_scenario_context
        self.scene_monologue_context = scene_monologue_context
        self.user_id = user_id
        self.scene_content = scene_content
        self.story_board_summary_context = story_board_summary_context
        self.evolutionary_step = evolutionary_step
        self.dreams_guidance_context = dreams_guidance_context
        self.dreams_personality_context = dreams_personality_context
        self.ref_analysis_id = ref_analysis_id
    @classmethod
    def from_config(cls, cfg=None):
        if cfg is None:
            raise RuntimeError("from_config cfg is None.")
        story_scenario_context = cfg.get("story_scenario_context", "")
        scene_monologue_context = cfg.get("scene_monologue_context", "")
        user_id = cfg.get("user_id", "")
        scene_content = cfg.get("scene_content", "")
        story_board_summary_context = cfg.get("story_board_summary_context", "")
        evolutionary_step = cfg.get("evolutionary_step", "")
        dreams_guidance_context = cfg.get("dreams_guidance_context", "")
        dreams_personality_context = cfg.get("dreams_personality_context", "")
        ref_analysis_id = cfg.get("ref_analysis_id", "")
        return cls(
            story_scenario_context=story_scenario_context,
            scene_monologue_context=scene_monologue_context,
            user_id=user_id,
            scene_content=scene_content,
            story_board_summary_context=story_board_summary_context,
            evolutionary_step=evolutionary_step,
            dreams_guidance_context=dreams_guidance_context,
            dreams_personality_context=dreams_personality_context,
            ref_analysis_id=ref_analysis_id
        )

    @classmethod
    def get_type(cls) -> str:
        """Get Object type."""
        return "dreams_node"

    def class_name(self) -> str:
        """Get class name."""
        if hasattr(self, "__class_getitem__"):
            return self.__class__.__name__
        elif hasattr(self, "__orig_class__"):
            return self.__orig_class__.__name__
        elif hasattr(self, "__name__"):
            return self.__name__
        else:
            return "心里分析信息节点"

    @property
    def template_content(self) -> str:
        return f"story_scenario_context: {self.story_scenario_context}, " \
               f"scene_monologue_context: {self.scene_monologue_context}, " \
               f"user_id: {self.user_id}, " \
               f"scene_content: {self.scene_content}, " \
               f"story_board_summary_context: {self.story_board_summary_context}, " \
               f"dreams_guidance_context: {self.dreams_guidance_context}, " \
               f"evolutionary_step: {self.evolutionary_step}, " \
               f"dreams_personality_context: {self.dreams_personality_context}, " \
               f"ref_analysis_id: {self.ref_analysis_id}"

    @template_content.setter
    def template_content(self, _template_content) -> None:
        raise RuntimeError("template_content is read only.")



import logging

import langchain

from dreamsboard.dreams.coplay_analysis_md.prompts import COSPLAY_ANALYSIS_MD_TEMPLATE
from dreamsboard.dreams.dreams_personality_chain.prompts import DREAMS_GEN_TEMPLATE
from dreamsboard.engine.storage.storage_context import StorageContext

from langchain_core.runnables import RunnableParallel, RunnableLambda
from langchain_core.prompt_values import StringPromptValue

logger = logging.getLogger(__name__)

_PROMPT_TEMPLATE_1 = langchain.PromptTemplate(input_variables=[
    "story_board_summary_context",
    "story_scenario_context",
    "scene_monologue_context",
    "user_id"],
    template=DREAMS_GEN_TEMPLATE)
_PROMPT_TEMPLATE_2 = langchain.PromptTemplate(
    input_variables=[
        "cosplay_role",
        "metadata",
        "story_scenario_context",
        "scene_monologue_context",
        "dreams_gen_text",
        "dreams_guidance_context",
        "evolutionary_step",
        "dreams_personality_context",
    ],
    template=COSPLAY_ANALYSIS_MD_TEMPLATE)


class CosplayAnalysisMD:

    def __init__(self,
                 cosplay_role: str,
                 source_url: str,
                 keyframe: str,
                 keyframe_path: str,
                 storage_keyframe: str,
                 storage_keyframe_path: str
                 ):
        self.cosplay_role = cosplay_role
        self.source_url = source_url
        self.keyframe = keyframe
        self.keyframe_path = keyframe_path
        self.storage_keyframe = storage_keyframe
        self.storage_keyframe_path = storage_keyframe_path
        self.storage_context = StorageContext.from_defaults(persist_dir=storage_keyframe_path)

    def format_md(self) -> StringPromptValue:
        def wrapper_guidance_unit(dict_input: dict):
            # 获取analysis_all第一个
            first_key = list(self.storage_context.dreams_analysis_store.analysis_all.keys())[0]
            analysis = self.storage_context.dreams_analysis_store.analysis_all[first_key]

            return analysis.to_dict()

        def wrapper_dreams_gen_unit(dict_input: StringPromptValue):
            # 获取analysis_all第一个
            first_key = list(self.storage_context.dreams_analysis_store.analysis_all.keys())[0]
            analysis = self.storage_context.dreams_analysis_store.analysis_all[first_key]

            return {
                "cosplay_role": self.cosplay_role,
                "dreams_gen_text": dict_input.text,
                "source_url": self.source_url,
                "keyframe": self.keyframe,
                "keyframe_path": self.keyframe_path,
                "storage_keyframe": self.storage_keyframe,
                "storage_keyframe_path": self.storage_keyframe_path,
                **analysis.to_dict()
            }

        chain = (RunnableLambda(wrapper_guidance_unit)
                 | _PROMPT_TEMPLATE_1
                 | RunnableLambda(wrapper_dreams_gen_unit)
                 | _PROMPT_TEMPLATE_2)

        out = chain.invoke({})
        return out

    def write_md(self, output_path: str) -> StringPromptValue:
        md = self.format_md()
        with open(output_path, "w") as f:
            f.write(md.text)
        logger.info(f"Write MD to {output_path}")
        return md



from __future__ import annotations
from abc import ABC
from typing import Any, Dict
from langchain_core.messages import ( 
    BaseMessage,
)
from langchain_core.language_models import LanguageModelInput
from langchain_core.runnables import Runnable
from langchain.chains.base import Chain
from langchain_core.prompts  import PromptTemplate
from langchain.chains import SequentialChain
from langchain_core.output_parsers import StrOutputParser
from langchain_core.runnables import RunnableParallel, RunnableLambda

from dreamsboard.dreams.dreams_personality_chain.prompts import (
    EDREAMS_PERSONALITY_TEMPLATE,
    STORY_BOARD_SCENE_TEMPLATE,
    STORY_BOARD_SUMMARY_CONTEXT_TEMPLATE,
    EDREAMS_EVOLUTIONARY_TEMPLATE,
    DREAMS_GEN_TEMPLATE,
)
from dreamsboard.document_loaders import StructuredStoryboardCSVBuilder
import os
import logging


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# 控制台打印
handler = logging.StreamHandler()
handler.setLevel(logging.INFO)

logger.addHandler(handler)


class StoryBoardDreamsGenerationChain(ABC):
    builder: StructuredStoryboardCSVBuilder
    dreams_guidance_chain: Chain
    dreams_personality_chain: Chain

    def __init__(self,
                 csv_file_path: str,
                 user_id: str,
                 dreams_guidance_chain: Chain,
                 dreams_personality_chain: Chain):
        self.builder = StructuredStoryboardCSVBuilder.form_builder(csv_file_path=csv_file_path)
        self.user_id = user_id
        self.dreams_guidance_chain = dreams_guidance_chain
        self.dreams_personality_chain = dreams_personality_chain

    @classmethod
    def from_dreams_personality_chain(
            cls,
            llm_runable: Runnable[LanguageModelInput, BaseMessage],
            csv_file_path: str,
            user_id: str = None,
    ) -> StoryBoardDreamsGenerationChain:
        # 03- 故事情境生成 `story_scenario_context`.txt STORY_BOARD_SCENE_TEMPLATE_Chain
        prompt_template1 = PromptTemplate(input_variables=["scene_content"],
                                          template=STORY_BOARD_SCENE_TEMPLATE)

        review_chain1 = (prompt_template1 | llm_runable | StrOutputParser())

        # 03-故事场景生成 `scene_monologue_context`.txt
        prompt_template2 = PromptTemplate(input_variables=["story_board_summary_context",
                                                           "user_id"],
                                          template=os.environ.get(
                                "STORY_BOARD_SUMMARY_CONTEXT_TEMPLATE", STORY_BOARD_SUMMARY_CONTEXT_TEMPLATE
                            ))
        review_chain2 = (prompt_template2 | llm_runable | StrOutputParser())
        # 04-情感情景引导.txt
        prompt_template = PromptTemplate(input_variables=["story_board_summary_context",
                                                          "story_scenario_context",
                                                          "scene_monologue_context",
                                                          "user_id"],
                                         template=os.environ.get(
                                "DREAMS_GEN_TEMPLATE", DREAMS_GEN_TEMPLATE
                            ))
        social_chain = (prompt_template | llm_runable | StrOutputParser())

        def wrapper_guidance_unit(_dict):
            return {
                # 中间变量包装至下一个pipline
                "story_scenario_context": _dict["story_scenario_context"],
                "scene_monologue_context": _dict["scene_monologue_context"],
                "user_id": _dict["user_id"],
                "scene_content": _dict["scene_content"],
                "story_board_summary_context": _dict["story_board_summary_context"],
            }

        def wrapper_guidance_output(_dict):
            return {
                # 中间变量全部打包输出
                "dreams_guidance_context": _dict["dreams_guidance_context"],
                "story_scenario_context": _dict["story_scenario_context"],
                "scene_monologue_context": _dict["scene_monologue_context"],
                "user_id": _dict["user_id"],
                "scene_content": _dict["scene_content"],
                "story_board_summary_context": _dict["story_board_summary_context"],
            }

        dreams_guidance_chain = ({
                                     "story_scenario_context": review_chain1,
                                     "scene_monologue_context": review_chain2,
                                     "user_id": lambda x: x["user_id"],
                                     "scene_content": lambda x: x["scene_content"],
                                     "story_board_summary_context": lambda x: x["story_board_summary_context"],
                                 } | RunnableLambda(wrapper_guidance_unit)
                                 | {
                                     "dreams_guidance_context": social_chain,
                                     "story_scenario_context": lambda x: x["story_scenario_context"],
                                     "scene_monologue_context": lambda x: x["scene_monologue_context"],
                                     "user_id": lambda x: x["user_id"],
                                     "scene_content": lambda x: x["scene_content"],
                                     "story_board_summary_context": lambda x: x["story_board_summary_context"],
                                 }
                                 | RunnableLambda(wrapper_guidance_output))

        # 05-剧情总结 `evolutionary_step`.txt
        prompt_template04 = PromptTemplate(input_variables=["story_board_summary_context"],
                                           template=os.environ.get(
                                "EDREAMS_EVOLUTIONARY_TEMPLATE", EDREAMS_EVOLUTIONARY_TEMPLATE
                            ))
        evolutionary_chain = prompt_template04 | llm_runable | StrOutputParser()
        # 05-性格分析 `dreams_personality_context`.txt
        prompt_template05 = PromptTemplate(input_variables=["evolutionary_step"],
                                           template=os.environ.get(
                                "EDREAMS_PERSONALITY_TEMPLATE", EDREAMS_PERSONALITY_TEMPLATE
                            ))
        personality_chain = prompt_template05 | llm_runable | StrOutputParser()

        def wrapper_personality_unit(_dict):
            return {
                # 中间变量包装至下一个pipline
                "evolutionary_step": _dict["evolutionary_step"],
                "story_board_summary_context": _dict["story_board_summary_context"],
            }

        def wrapper_personality_output(_dict):
            return {
                # 中间变量全部打包输出
                "dreams_personality_context": _dict["dreams_personality_context"],
                "evolutionary_step": _dict["evolutionary_step"],
                "story_board_summary_context": _dict["story_board_summary_context"],
            }

        dreams_personality_chain = ({
                                        "story_board_summary_context": lambda x: x["story_board_summary_context"],
                                        "evolutionary_step": evolutionary_chain
                                    } | RunnableLambda(wrapper_personality_unit)
                                    | {
                                        "dreams_personality_context": personality_chain,
                                        "evolutionary_step": lambda x: x["evolutionary_step"],
                                        "story_board_summary_context": lambda x: x["story_board_summary_context"],
                                    }
                                    | RunnableLambda(wrapper_personality_output))

        return cls(csv_file_path=csv_file_path,
                   user_id=user_id if user_id else "此来访者",
                   dreams_guidance_chain=dreams_guidance_chain,
                   dreams_personality_chain=dreams_personality_chain)

    def run(self) -> Dict[str, Any]:
        # 对传入的剧本台词转换成 scene_content
        self.builder.load()
        selected_columns = ["story_board_role", "story_board_text", "story_board"]
        scene_content = self.builder.build_text(self.user_id, selected_columns)
        story_board_summary_context = self.builder.build_msg()

        dreams_guidance_personality_map_chain = RunnableParallel(
            dreams_guidance_context=self.dreams_guidance_chain,
            dreams_personality_context=self.dreams_personality_chain)

        return dreams_guidance_personality_map_chain.invoke(
            {"scene_content": scene_content,
             "story_board_summary_context": story_board_summary_context,
             "user_id": self.user_id})
    
    def invoke_dreams_guidance_chain(self, input: Dict[str, Any]) -> Dict[str, Any]:
        return self.dreams_guidance_chain.invoke(input)
    
    def invoke_dreams_personality_chain(self, input: Dict[str, Any]) -> Dict[str, Any]:
        return self.dreams_personality_chain.invoke(input)



from langchain.schema import BaseMessage

from dreamsboard.engine.data_structs.data_structs import IndexStruct, IndexDict
from dreamsboard.engine.schema import BaseNode
from dreamsboard.engine.storage.storage_context import StorageContext
from dreamsboard.engine.generate.code_executor import CodeExecutor
from dreamsboard.engine.generate.code_generate import CodeGenerator
from dreamsboard.engine.generate.run_generate import CodeGeneratorHandler, CodeGeneratorChain
from typing import Any, Dict, List, Generic, TypeVar, Sequence, Type, Optional

from langchain_core.messages import ( 
    BaseMessage,
)
from langchain_core.language_models import LanguageModelInput
from langchain_core.runnables import Runnable

from langchain.schema.language_model import BaseLanguageModel
from abc import ABC, abstractmethod

from dreamsboard.engine.storage.template_store.types import BaseTemplateStore

IS = TypeVar("IS", bound=IndexStruct)
EngineBuilderType = TypeVar("EngineBuilderType", bound="BaseEngineBuilder")


class BaseEngineBuilder(Generic[IS], ABC):

    index_struct_cls: Type[IS]
    _index_struct: Optional[IS] = None

    def __init__(
            self,
            nodes: Optional[Sequence[CodeGenerator]] = None,
            storage_context: Optional[StorageContext] = None,
            index_struct: Optional[IS] = None,
            show_progress: bool = False,
            **kwargs: Any,
    ) -> None:
        """Initialize with parameters."""
        if index_struct is None and nodes is None:
            raise ValueError("One of nodes or index_struct must be provided.")
        if index_struct is not None and nodes is not None:
            raise ValueError("Only one of nodes or index_struct can be provided.")
        # This is to explicitly make sure that the old UX is not used
        if nodes is not None and len(nodes) >= 1 and not isinstance(nodes[0], CodeGenerator):
            raise ValueError("nodes must be a list of CodeGenerator objects.")
        self._storage_context = storage_context or StorageContext.from_defaults()
        self._template_store = self._storage_context.template_store
        self._show_progress = show_progress

        if index_struct is None:
            assert nodes is not None
            for node in nodes:
                self._template_store.set_template_hash(node.node_id, node.hash)
            self._index_struct = self.build_index_from_nodes(nodes)

        elif storage_context is not None and nodes is None and index_struct is not None:
            for node_id, node in storage_context.template_store.templates.items():
                if node is not None:
                    self._template_store.set_template_hash(node_id, node.hash)

            self._index_struct = index_struct
            self.build_index_from_nodes(storage_context.template_store.templates.values())

        self._storage_context.index_store.add_index_struct(self._index_struct)

    @classmethod
    def from_template(
            cls: Type[EngineBuilderType],
            nodes: Optional[Sequence[CodeGenerator]] = None,
            storage_context: Optional[StorageContext] = None,
            show_progress: bool = False,
            **kwargs: Any,
    ) -> EngineBuilderType:
        """Create index from documents.

        Args:

        """
        storage_context = storage_context or StorageContext.from_defaults()
        return cls(
            nodes=nodes,
            storage_context=storage_context,
            show_progress=show_progress,
            **kwargs,
        )

    @property
    def index_struct(self) -> IS:
        """Get the index struct."""
        return self._index_struct

    @property
    def index_id(self) -> str:
        """Get the index struct."""
        return self._index_struct.index_id

    def set_index_id(self, index_id: str) -> None:
        """Set the index id.

        NOTE: if you decide to set the index_id on the index_struct manually,
        you will need to explicitly call `add_index_struct` on the `index_store`
        to update the index store.

        .. code-block:: python
            index.index_struct.index_id = index_id
            index.storage_context.index_store.add_index_struct(index.index_struct)

        Args:
            index_id (str): Index id to set.

        """
        # delete the old index struct
        old_id = self._index_struct.index_id
        self._storage_context.index_store.delete_index_struct(old_id)
        # add the new index struct
        self._index_struct.index_id = index_id
        self._storage_context.index_store.add_index_struct(self._index_struct)

    @property
    def template_store(self) -> BaseTemplateStore:
        """Get the _template_store corresponding to the index."""
        return self._template_store

    @property
    def storage_context(self) -> StorageContext:
        return self._storage_context

    @property
    def summary(self) -> str:
        return str(self._index_struct.summary)

    @summary.setter
    def summary(self, new_summary: str) -> None:
        self._index_struct.summary = new_summary
        self._storage_context.index_store.add_index_struct(self._index_struct)

    @abstractmethod
    def _build_index_from_nodes(self, nodes: Sequence[CodeGenerator]) -> IS:
        """Build the index from nodes."""

    def build_index_from_nodes(self, nodes: Sequence[CodeGenerator]) -> IS:
        """Build the index from nodes."""
        self._template_store.add_templates(nodes, allow_update=True)
        return self._build_index_from_nodes(nodes)

    @abstractmethod
    def _add_generator(self, nodes: Sequence[CodeGenerator], **insert_kwargs: Any) -> None:
        """Index-specific logic for inserting nodes to the index struct."""

    def add_generators(self, nodes: Sequence[CodeGenerator], **insert_kwargs: Any) -> None:
        self._template_store.add_templates(nodes, allow_update=True)
        self._add_generator(nodes, **insert_kwargs)
        self._storage_context.index_store.add_index_struct(self._index_struct)

    def add_generator(self, node: CodeGenerator, **insert_kwargs: Any) -> None:
        """Insert a document."""

        self.add_generators([node], **insert_kwargs)

        self._template_store.set_template_hash(node.node_id, node.hash)


# Create a code generator builder
class CodeGeneratorBuilder(BaseEngineBuilder[IndexDict]):
    """
    BaseEngineBuilder[IndexDict]
    BaseEngineBuilder: 构建器基类，用于构建代码生成器的构建器.
    IndexDict: 用于存储索引的数据结构,支持链式调用、序列化和反序列化
    code_gen_chain: 代码链接器，用于存储代码生成器
    用于构建代码生成器的构建器，支持链式调用、序列化和反序列化
    被链接起来的代码生成器会按照添加的顺序依次执行，生成最终的代码
    连接器可被序列化，用于保存和加载
    """
    index_struct_cls = IndexDict
    _code_gen_chain: CodeGeneratorChain = CodeGeneratorChain()

    def _build_index_from_nodes(self, nodes: Sequence[CodeGenerator]) -> IndexDict:
        """Build index from nodes."""
        if self._index_struct is None:
            index_struct = self.index_struct_cls()
        else:
            index_struct = self._index_struct
        self._add_nodes_to_index(
            index_struct, nodes, show_progress=self._show_progress
        )

        return index_struct

    def _add_generator(self, nodes: Sequence[CodeGenerator], **insert_kwargs: Any) -> None:

        self._add_nodes_to_index(self._index_struct, nodes, **insert_kwargs)

    def _add_nodes_to_index(
            self,
            index_struct: IndexDict,
            nodes: Sequence[CodeGenerator],
            show_progress: bool = False,
    ) -> None:
        """
        添加节点到索引中，节点的
        :param index_struct:
        :param nodes:
        :param show_progress:
        :return:
        """
        if not nodes:
            return

        for node in nodes:

            self._code_gen_chain.add_generator(node)
            # NOTE: remove embedding from node to avoid duplication
            node_without_copy = node.copy()

            index_struct.add_node(node_without_copy)

    def remove_last_generator(self):
        # 先删除缓存中的代码生成器
        self._index_struct.delete(doc_id=self._code_gen_chain.chain_tail.generator.node_id)
        self._template_store.delete_template(self._code_gen_chain.chain_tail.generator.node_id, raise_error=False)
        self._code_gen_chain.remove_last_generator()

    def build_executor(self, llm_runable: Runnable[LanguageModelInput, BaseMessage], messages: List[BaseMessage] = [], render_data: dict = {}) -> CodeExecutor:
        if self._code_gen_chain.chain_head is None:
            raise RuntimeError("chain_head is None.")

        executor_code = self._code_gen_chain.generate(render_data)
        self.summary = executor_code
        executor = CodeExecutor(executor_code=executor_code, llm_runable=llm_runable, messages=messages)
        return executor




from langchain_core.tools import BaseTool
from langchain_core.utils.function_calling import convert_to_openai_tool
from langchain_core.pydantic_v1 import BaseModel
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
)
from langchain_core.tools import BaseTool
from dreamsboard.common.struct_type import AdapterAllToolStructType



def _is_assistants_builtin_tool(
    tool: Union[Dict[str, Any], Type[BaseModel], Callable, BaseTool],
) -> bool:
    """platform tools built-in"""
    assistants_builtin_tools = AdapterAllToolStructType.__members__.values()
    return (
        isinstance(tool, dict)
        and ("type" in tool)
        and (tool["type"] in assistants_builtin_tools)
    )



def _get_assistants_tool(
    tool: Union[Dict[str, Any], Type[BaseModel], Callable, BaseTool],
) -> Dict[str, Any]:
    """Convert a raw function/class to an ZhipuAI tool."""
    if _is_assistants_builtin_tool(tool):
        return tool  # type: ignore
    else:
        # in case of a custom tool, convert it to an function of type
        return convert_to_openai_tool(tool)
